from orgasm.indexer._orgasm cimport *
from orgasm.graph._graph cimport *

import sys

cdef tuple findDeepestRead(Index index, bytes seed):
    cdef list frg
    cdef list fcount
    cdef int mfc
    
    frg   = [seed[x:(x+index.getReadSize())] for x in xrange(len(seed)-index.getReadSize())]
    fcount= [index.count(x) for x in frg]
    mfc = max(fcount)
    return frg[fcount.index(mfc)],mfc


cdef class Assembler:

    cdef DiGraph _graph
    cdef Index _index
    cdef int _overlap
    cdef set extensionPoints

    def __init__(self,Index index, bytes seed, int overlap=90):
        '''
        Create a new assembler object.
        
        :param index: the file name of an read index previously formated using orgasmi binary
        :type index: bytes
        :param seed: a DNA sequenced used as started for the assembly
        :type seed: bytes
        :param overlap: the minimum overlap between two reads
        :type overlap: int
        '''
        cdef bytes starter
        cdef int depth 
        cdef tuple reads
        cdef dict attr
        
        (starter,depth) = findDeepestRead(index,seed)
        
        self._graph = DiGraph(b"assembling")
        
        self._index = index
        self._overlap=overlap
        
        reads = self.readType([x for x in index[starter]])
        
        self._graph.addNode(reads[0])
        attr = self._graph[reads[0]]
        attr[b"weight"]=reads[1]
        attr[b"reads"]=reads[2]
        
        self.extensionPoints = set([(starter,reads[0])])
        
    cpdef tuple readType(self,ids):
        '''
        Internal function : Given a set of read ids, return the one that have to be used
        as standard id. 
        
        The set of ids given to this function corresponds to a set of all strictly identical reads.
         
        :param ids: an iterable elements contining read ids.
        :type ids: iterable
        
        :returns: a tuble of three elements
        
                    - the standard id to use
                    - the length of ids
                    - a set containing all unique ids given as parameter
                    
        :rtype: tuple
        '''
        
        cdef int t 
        cdef int y 
        
        if not ids:
            return None
        
        t = len(self._index)+1
        
        for y in ids:
            if abs(y) < abs(t):
                t = y
                
        return (t,len(ids),set(ids))
    
    cpdef run(self, int minweight=1):
        cdef int cycle
        cdef int extcount
    
        cycle = 0
        extcount = len(self.extensionPoints)
        
        while self.extensionPoints and extcount < 1000 :
            cycle+=1
            extcount = len(self.extensionPoints)
            print >>sys.stderr,b"Cycle : %8d     Extension point : %4d \r" % (cycle,extcount),
            self.runOneCycle(minweight)
    
    cpdef runOneCycle(self, int minweight=1):
        cdef set extensions = set()
        cdef tuple e
        
        for e in self.extensionPoints:
            extensions|=self.nextExtention(e[0],e[1],minweight)

        self.extensionPoints=extensions
        
    def  nextExtention(self,seed,nodeindex, int minweight=1):
        cdef int   i
        cdef dict  e
        cdef bytes s 
        cdef set  starter
        cdef bytes   ex
        cdef tuple reads 
        cdef int   nodeindexE
        cdef dict  attr
        cdef set   son
        
        i = self._index.getReadSize() - 1;
        
        e = {}
        
        while i >=self._overlap and not e:
            s = seed[-i:]
            e = self._index.getExtensions(s)
            i-=1
            
        starter = set()
        for ex in e:
            data = e[ex]
            reads = self.readType(data[1])
            if reads[1] >= minweight:
                nodeindexE = reads[0]
                
                self._graph.addEdge(nodeindex, nodeindexE)
                
                attr = self._graph.getNodeAttr(nodeindexE)
                attr[b"weight"]=reads[1]
                attr[b"reads"]=reads[2]
                
                attr = self._graph.getEdgeAttr(nodeindex, nodeindexE)
                attr[b"label"]=ex
                attr[b"arrowsize"]=float(reads[1])/10
                attr[b"type"]=b"next read"
                
                
                son = set(self._graph.neighbourIterator(nodeindexE,edgePredicate=lambda e : self._graph.getEdgeAttr(e[0],e[1])[b'type']==b"next read"))
                if not son:
                    starter.add((self._index.getRead(nodeindexE,0,self._index.getReadSize()),nodeindexE))            
            
        return starter
    
    def forkIterator(self):
        
        def has_several_sons(int node):
            return  len(set(self._graph.neighbourIterator(node,edgePredicate=lambda e : self._graph.getEdgeAttr(e[0],e[1])[b'type']==b"next read"))) > 1
        
        return self._graph.nodeIterator(has_several_sons)
    

        
    def deepMicroBubble(self,int startnode,int lenmax=-1, int wide=100):
        def sons(int node):
            return set(self._graph.neighbourIterator(node,edgePredicate=lambda e : self._graph.getEdgeAttr(e[0],e[1])[b'type']==b"next read"))
        
        def intersect(list runs):
            return reduce(lambda x,y: x & y,(set(r[3]) for r in runs if r[0]))
        
        if lenmax < 0:
            lenmax = self._index.getReadSize()
            
        graph = self._graph
            
        stack = [(0,[startnode])]
        refpath = []
        allpaths = []
        allnodes = set()

        while stack:
            length,path = stack.pop()
            first=0
            
            while length < lenmax and first not in refpath:
                brothers = sons(path[-1])
                first = brothers.pop()
                
                for b in brothers:
                    e = self.graph.getEdgeAttr(path[-1],b)
                    l = len(e[b'label']) + length
                    stack.append((l,path + [b]))
                    
                e = self.graph.getEdgeAttr(path[-1],first)
                path.append(first)
                length = len(e[b'label']) + length
            else:
                if first not in refpath:    
                    allpaths.append(path)

            allnodes |= set(path)
            
            if not refpath:
                refpath = set(path)
                                
        return allpaths,self.graph.subgraph("microbubble",allnodes)

                

    def microBubble(self,int startnode,int lenmax=-1, int wide=100):
        cdef set     brothers
        cdef DiGraph graph
        cdef list    runs
        cdef set     inter 
        cdef bint    newpath
        cdef list    newruns
        cdef bint    ok 
        cdef int     node
        cdef int     plength
        cdef list    path
        cdef list    extraruns
        cdef list    toappend
        cdef list    pathok
        cdef int     out
        cdef set     sout 
        cdef list    runmax     
        cdef int     cov 
        cdef int     s 
        cdef dict    n
        cdef list    goodpath
        cdef int     b 
        cdef set     nodeset
        cdef set     badset
        
        def sons(int node):
            return set(self._graph.neighbourIterator(node,edgePredicate=lambda e : self._graph.getEdgeAttr(e[0],e[1])[b'type']==b"next read"))
        
        def intersect(list runs):
            return reduce(lambda x,y: x & y,(set(r[3]) for r in runs if r[0]))
        
        def pathmax(list runs):
            return max(x[2] for x in runs if x[0])
        
        if lenmax < 0:
            lenmax = self._index.getReadSize()
            
        brothers = sons(startnode)
        graph = self._graph
            
        runs = [[True,
                 x,
                 len(graph.getEdgeAttr(startnode,x)[b'label']),
                 [x]] 
                 for x in brothers]
                
        inter = intersect(runs)
        newpath=True

        while newpath and len(inter)==0 and pathmax(runs) < lenmax:
            newruns=[]
            newpath=False
            for ok,node,plength,path in runs:
                if ok:
                    brothers = sons(node)
 #                   loop = reduce(lambda x,y: x or y, ((x in path) for x in brothers),False)
 #                   if loop:
 #                        print >> sys.stderr,"Loop : %s" % (path + [x])
                    extraruns = [[x not in path,x,
                                  plength + len(graph.getEdgeAttr(node,x)[b'label']),
                                  path + [x]
                                 ]
                                 for x in brothers
                                ]

                    if extraruns: 
                        toappend=[x for x in extraruns if x]
                        if toappend:         
                            newruns.extend(toappend)
                            newpath=True
                    else:
                        newruns.append([False,node,plength,path])
                else:
                    newruns.append([False,node,plength,path])
                    
            runs = newruns
                
            if len(runs) > wide:
#                print >> sys.stderr,"Microbubble to wide (%d)" % len(runs)
#                xxx=set()
#                for r in runs:
#                    xxx|=set(r[3])
#                xxx.add(startnode)
#                sg = self.graph.subgraph("toowide",xxx)
#                print >> sys.stderr,"Starting from %d" % startnode
#                print >>open(b'toowide.dot',b'w'),sg
#                print >>open("toowide.dat","w"),runs
#                sys.exit(1)
                    

                return None,0
            
            if newpath:
                inter = intersect(runs)
                        
        if newpath:
            if len(inter)==0:
                #print >> sys.stderr,"Microbubble to long (%d)" % pathmax(runs)
                return None,0
            
            pathok = [x[3] for x in runs if x[0]]
            
            sout = set(y[min(y.index(z) for z in inter)] for y in pathok)
    
            assert len(sout)==1,b"Bug in the bubble detection during analysis of the bubblue starting in %d" % startnode
            
            out = sout.pop()
            
            graph.getNodeAttr(out)[b'type']=b"microbubble out"
        
            for ok,node,plength,path in runs:
                try:
                    del path[path.index(out):]
                except ValueError:
                    pass
                
            runmax=[]
            
            for path in pathok:
                cov = 0
                for s in path:
                    n = graph.getNodeAttr(s)
                    cov+=n[b'weight']
                runmax.append(cov)    
                
            
            goodpath = pathok[runmax.index(max(runmax))]
            
            
            b = startnode
            graph.getNodeAttr(b)[b'color']=b"green"
        
            for e in goodpath:
                graph.getNodeAttr(e)[b'color']=b"green"
                graph.getEdgeAttr(b,e)[b'color']=b'green'
                b=e
                    
            graph.getEdgeAttr(b,out)[b'color']=b'green'
            graph.getNodeAttr(out)[b'color']=b"green"
        
            nodeset = set([startnode,out]) 
        
            for r in runs:
                nodeset |= set(r[3])
        
            badset=(nodeset - set(goodpath)) - set([startnode,out])

        else:
            graph.getNodeAttr(startnode)[b'color']=b"green"
            nodeset = set([startnode]) 
        
            for r in runs:
                nodeset |= set(r[3])
        
            badset=nodeset - set([startnode])
            out=0
            
        
        for bad in badset:
            badnode = graph.getNodeAttr(bad)
            badnode[b'color']=b"red"
            badnode[b'type']=b"to remove"
    
    
        return self._graph.subgraph(b'microbubble',nodeset),out
        
    def microBubbleIterator(self,int lenmax=-1, int wide=100):
        cdef DiGraph b
        cdef int f
        cdef int out
        cdef set bnode
        cdef NodeIterator ibubble
        
        ibubble = self.forkIterator()
        
        for f in ibubble:
            b,out = self.microBubble(f,lenmax,wide)
            if b is not None:
                bnode = set(b) 
                if out==0:
                    bnode.remove(out)
                ibubble|=bnode
                    
                yield b
                    
                            
    def resolveMicroBubble(self,int lenmax=-1, int wide=100):
        cdef int n = 0
        cdef DiGraph b
        cdef list todel
                
        print >>sys.stderr,b"Resolving micro bubbles"
        
        for b in self.microBubbleIterator(lenmax,wide):
            n+=1
            print >>sys.stderr,b"Bubble : %8d\r" %n,
            
            if b.nodeCount() > 40:
                print >>open(b'bubble.dot',b'w'),b
                print >>sys.stderr,b"Bubble : %8d >>> new graph %10s\r" % (n,b.nodeIterator().next()),
                
        print >>sys.stderr
        
        todel = [x for x in self._graph.nodeIterator(lambda n:self._graph.getNodeAttr(n).get(b"type",None)==b"to remove")]

        print >>sys.stderr,b"Cleanning %8d nodes" % len(todel)
        
        for n in todel:
            self._graph.deleteNode(n) 
            
    def endNodeSet(self):
        def isfinalnode(int node):
            return len(set(self._graph.neighbourIterator(node,edgePredicate=lambda e : self._graph.getEdgeAttr(e[0],e[1])[b'type']==b"next read")))==0
        return set(x for x in self._graph.nodeIterator(isfinalnode))

    def cleanDeadBranches(self):
        cdef int i
        cdef int p
        cdef set tocut
        cdef list nextcut
    
        def isfinalnode(int node):
            return len(set(self._graph.neighbourIterator(node,edgePredicate=lambda e : self._graph.getEdgeAttr(e[0],e[1])[b'type']==b"next read")))==0
        
        print >>sys.stderr,b"Cleanning dead branches"
        endpoints = self.endNodeSet()
        
        i=0
                
        for p in endpoints:
            tocut = set([p])
            while tocut:
                nextcut=[]
                for p in tocut:
                    nextcut.extend(self._graph.parentIterator(p))
                    self._graph.deleteNode(p)
                    i+=1
                tocut = set(p for p in nextcut if isfinalnode(p))
                
                print >>sys.stderr,"%9d deleted nodes\r" % i,
                    
                
        print >>sys.stderr
                
        return i
    
    cpdef simplifyAssembly(self,int lenmax=-1, int wide=100):
        cdef int l
               
        self.cleanDeadBranches()
        
        l = 0
        
        while (l!=len(self._graph)):
            l=self._graph.nodeCount()
            self.resolveMicroBubble(lenmax, wide)
            self.cleanDeadBranches()
            
    def markOneStepPath(self,int jump=1):
        cdef int n1
        cdef int n2

        def singlestep(e):
            return len(self._graph.getEdgeAttr(e[0],e[1]).get(b'label',''))==jump
        
        for n1,n2 in self._graph.edgeIterator(None,singlestep):
            self._graph.getNodeAttr(n1)[b'color']=b'blue'
            self._graph.getNodeAttr(n2)[b'color']=b'blue'
            self._graph.getEdgeAttr(n1,n2)[b'color']=b'blue'
            self._graph.getNodeAttr(n1)[b'singlestep']=True
            self._graph.getNodeAttr(n2)[b'singlestep']=True
            self._graph.getEdgeAttr(n1,n2)[b'singlestep']=True
            
    def findAlternativePath(self,set nodes):
        
        
        def walk(node,length=0):
            pass
        sg = self._graph.subDiGraph("single",nodes)
        start = [x for x in sg.nodeIterator(lambda n: len(set(sg.parentIterator(n,edgePredicate=lambda e:sg[e].get(b'color','')==b'blue')))==0)]
        
        
                
    def __len__(self):
        return self._graph.nodeCount()
    
    def stemIterator(self):
        cdef NodeIterator ni
        cdef int first
        cdef int n
        cdef set branches
        cdef set brothers
        cdef set fathers
        cdef list lsequence
        cdef int last
        cdef int lseq
        cdef bytes sequence
        
        def sons(int node):
            return set(self._graph.neighbourIterator(node,edgePredicate=lambda e : self._graph.getEdgeAttr(e[0],e[1])[b'type']==b"next read"))
            
        def parents(int node):
            return set(self._graph.parentIterator(node,edgePredicate=lambda e : self._graph.getEdgeAttr(e[0],e[1])[b'type']==b"next read"))

        def is_junction(int node):
            return  len(set(self._graph.neighbourIterator(node,edgePredicate=lambda e : self._graph.getEdgeAttr(e[0],e[1])[b'type']==b"next read"))) != 1 or \
                    len(set(self._graph.parentIterator(node,edgePredicate=lambda e : self._graph.getEdgeAttr(e[0],e[1])[b'type']==b"next read"))) != 1
        
        ni=self._graph.nodeIterator(is_junction)

        for first in ni:
            branches=sons(first)
            
            for n in branches:
                
                brothers=sons(n)
                fathers=parents(n)
                lsequence=[]
                
                lsequence.append(self._graph.getEdgeAttr(first,n)[b'label'])
                                
                while len(brothers)==1 and len(fathers)==1:
                    son=brothers.pop()
                    lsequence.append(self._graph.getEdgeAttr(n,son)[b'label'])
                    n=son
                    brothers=sons(n)
                    fathers=parents(n)
    
                last = n
                sequence=b''.join(lsequence)
                lseq = len(sequence)
                
                yield first,last,lseq,sequence

        
    cpdef DiGraph compactAssembling(self):
        cdef DiGraph graph = DiGraph('compact')
        cdef int n=0
        cdef lcontig=0
        cdef int first 
        cdef int last
        cdef int lseq
        cdef bytes sequence
        cdef dict attr

        print >>sys.stderr,b"Compacting graph :"
        
        for first,last,lseq,sequence in self.stemIterator():
            n+=1
            lcontig+=lseq
            print >>sys.stderr,b" Stem %9d : %6d bp (total : %6d)" % (n,lseq,lcontig)
            graph.addEdge(first,last)
            attr = graph.getEdgeAttr(first,last)
            attr[b'sequence']=sequence
            attr[b'length']=lseq
            if lseq > 10:
                attr[b'label']=b"%s->(%d)->%s" % (sequence[0:5],lseq,sequence[-5:])
            else:
                attr[b'label']=b"%s->(%d)" % (sequence,lseq)
                
        return graph
    
    property graph:

        "A doc string can go here."

        def __get__(self):
            return self._graph


                       
